@@ -3,7 +3,7 @@ use std::{
     sync::atomic::{AtomicBool, Ordering},
 };
 
-use gix_odb::{store::RefreshMode, FindExt};
+use gix_odb::store::RefreshMode;
 use gix_protocol::{
     fetch::Arguments,
     transport::{client::Transport, packetline::read::ProgressAction},
@@ -23,14 +23,14 @@ use crate::{
             Shallow, Status,
         },
     },
-    Progress, Repository,
+    Repository,
 };
 
 impl<'remote, 'repo, T> Prepare<'remote, 'repo, T>
 where
     T: Transport,
 {
-    /// Receive the pack and perform the operation as configured by git via `gix-config` or overridden by various builder methods.
+    /// Receive the pack and perform the operation as configured by git via `git-config` or overridden by various builder methods.
     /// Return `Ok(None)` if there was nothing to do because all remote refs are at the same state as they are locally, or `Ok(Some(outcome))`
     /// to inform about all the changes that were made.
     ///
@@ -73,11 +73,21 @@ where
     /// - `gitoxide.userAgent` is read to obtain the application user agent for git servers and for HTTP servers as well.
     ///
     #[gix_protocol::maybe_async::maybe_async]
-    pub async fn receive<P>(mut self, mut progress: P, should_interrupt: &AtomicBool) -> Result<Outcome, Error>
+    pub async fn receive<P>(self, mut progress: P, should_interrupt: &AtomicBool) -> Result<Outcome, Error>
     where
-        P: Progress,
+        P: gix_features::progress::NestedProgress,
         P::SubProgress: 'static,
     {
+        self.receive_inner(&mut progress, should_interrupt).await
+    }
+
+    #[gix_protocol::maybe_async::maybe_async]
+    #[allow(clippy::drop_non_drop)]
+    pub(crate) async fn receive_inner(
+        mut self,
+        progress: &mut dyn crate::DynNestedProgress,
+        should_interrupt: &AtomicBool,
+    ) -> Result<Outcome, Error> {
         let _span = gix_trace::coarse!("fetch::Prepare::receive()");
         let mut con = self.con.take().expect("receive() can only be called once");
 
@@ -85,7 +95,6 @@ where
         let protocol_version = handshake.server_protocol_version;
 
         let fetch = gix_protocol::Command::Fetch;
-        let progress = &mut progress;
         let repo = con.remote.repo;
         let fetch_features = {
             let mut f = fetch.default_features(protocol_version, &handshake.capabilities);
@@ -95,7 +104,7 @@ where
 
         gix_protocol::fetch::Response::check_required_features(protocol_version, &fetch_features)?;
         let sideband_all = fetch_features.iter().any(|(n, _)| *n == "sideband-all");
-        let mut arguments = gix_protocol::fetch::Arguments::new(protocol_version, fetch_features);
+        let mut arguments = gix_protocol::fetch::Arguments::new(protocol_version, fetch_features, con.trace);
         if matches!(con.remote.fetch_tags, crate::remote::fetch::Tags::Included) {
             if !arguments.can_use_include_tag() {
                 return Err(Error::MissingServerFeature {
@@ -116,7 +125,10 @@ where
             });
         }
 
-        let negotiate_span = gix_trace::detail!("negotiate");
+        let negotiate_span = gix_trace::detail!(
+            "negotiate",
+            protocol_version = self.ref_map.handshake.server_protocol_version as usize
+        );
         let mut negotiator = repo
             .config
             .resolved
@@ -146,7 +158,9 @@ where
         let mut previous_response = None::<gix_protocol::fetch::Response>;
         let (mut write_pack_bundle, negotiate) = match &action {
             negotiate::Action::NoChange | negotiate::Action::SkipToRefUpdate => {
-                gix_protocol::indicate_end_of_interaction(&mut con.transport).await.ok();
+                gix_protocol::indicate_end_of_interaction(&mut con.transport, con.trace)
+                    .await
+                    .ok();
                 (None, None)
             }
             negotiate::Action::MustNegotiate {
@@ -167,7 +181,7 @@ where
                 let mut seen_ack = false;
                 let mut in_vain = 0;
                 let mut common = is_stateless.then(Vec::new);
-                let reader = 'negotiation: loop {
+                let mut reader = 'negotiation: loop {
                     let _round = gix_trace::detail!("negotiate round", round = rounds.len() + 1);
                     progress.step();
                     progress.set_name(format!("negotiate (round {})", rounds.len() + 1));
@@ -193,11 +207,13 @@ where
                                 previous_response_had_at_least_one_in_common: ack_seen,
                             });
                             let is_done = haves_sent != haves_to_send || (seen_ack && in_vain >= 256);
-                            haves_to_send = gix_negotiate::window_size(is_stateless, haves_to_send);
+                            haves_to_send = gix_negotiate::window_size(is_stateless, Some(haves_to_send));
                             is_done
                         }
                         Err(err) => {
-                            gix_protocol::indicate_end_of_interaction(&mut con.transport).await.ok();
+                            gix_protocol::indicate_end_of_interaction(&mut con.transport, con.trace)
+                                .await
+                                .ok();
                             return Err(err.into());
                         }
                     };
@@ -205,13 +221,18 @@ where
                     if sideband_all {
                         setup_remote_progress(progress, &mut reader, should_interrupt);
                     }
-                    let response =
-                        gix_protocol::fetch::Response::from_line_reader(protocol_version, &mut reader, is_done).await?;
+                    let response = gix_protocol::fetch::Response::from_line_reader(
+                        protocol_version,
+                        &mut reader,
+                        is_done,
+                        !is_done,
+                    )
+                    .await?;
                     let has_pack = response.has_pack();
                     previous_response = Some(response);
                     if has_pack {
                         progress.step();
-                        progress.set_name("receiving pack");
+                        progress.set_name("receiving pack".into());
                         if !sideband_all {
                             setup_remote_progress(progress, &mut reader, should_interrupt);
                         }
@@ -245,31 +266,48 @@ where
                 };
 
                 let write_pack_bundle = if matches!(self.dry_run, fetch::DryRun::No) {
-                    Some(gix_pack::Bundle::write_to_directory(
-                        #[cfg(feature = "async-network-client")]
-                        {
-                            gix_protocol::futures_lite::io::BlockOn::new(reader)
-                        },
-                        #[cfg(not(feature = "async-network-client"))]
-                        {
-                            reader
-                        },
-                        Some(repo.objects.store_ref().path().join("pack")),
+                    #[cfg(not(feature = "async-network-client"))]
+                    let mut rd = reader;
+                    #[cfg(feature = "async-network-client")]
+                    let mut rd = gix_protocol::futures_lite::io::BlockOn::new(reader);
+                    let res = gix_pack::Bundle::write_to_directory(
+                        &mut rd,
+                        Some(&repo.objects.store_ref().path().join("pack")),
                         progress,
                         should_interrupt,
                         Some(Box::new({
                             let repo = repo.clone();
-                            move |oid, buf| repo.objects.find(oid, buf).ok()
+                            repo.objects
                         })),
                         options,
-                    )?)
+                    )?;
+                    // Assure the final flush packet is consumed.
+                    #[cfg(feature = "async-network-client")]
+                    let has_read_to_end = { rd.get_ref().stopped_at().is_some() };
+                    #[cfg(not(feature = "async-network-client"))]
+                    let has_read_to_end = { rd.stopped_at().is_some() };
+                    if !has_read_to_end {
+                        std::io::copy(&mut rd, &mut std::io::sink()).map_err(Error::ReadRemainingBytes)?;
+                    }
+                    #[cfg(feature = "async-network-client")]
+                    {
+                        reader = rd.into_inner();
+                    }
+
+                    #[cfg(not(feature = "async-network-client"))]
+                    {
+                        reader = rd;
+                    }
+                    Some(res)
                 } else {
-                    drop(reader);
                     None
                 };
+                drop(reader);
 
                 if matches!(protocol_version, gix_protocol::transport::Protocol::V2) {
-                    gix_protocol::indicate_end_of_interaction(&mut con.transport).await.ok();
+                    gix_protocol::indicate_end_of_interaction(&mut con.transport, con.trace)
+                        .await
+                        .ok();
                 }
 
                 if let Some(shallow_lock) = shallow_lock {
@@ -372,27 +410,14 @@ fn add_shallow_args(
     Ok((shallow_commits, shallow_lock))
 }
 
-fn setup_remote_progress<P>(
-    progress: &mut P,
-    reader: &mut Box<dyn gix_protocol::transport::client::ExtendedBufRead + Unpin + '_>,
-    should_interrupt: &AtomicBool,
-) where
-    P: Progress,
-    P::SubProgress: 'static,
-{
+fn setup_remote_progress<'a>(
+    progress: &mut dyn crate::DynNestedProgress,
+    reader: &mut Box<dyn gix_protocol::transport::client::ExtendedBufRead<'a> + Unpin + 'a>,
+    should_interrupt: &'a AtomicBool,
+) {
     use gix_protocol::transport::client::ExtendedBufRead;
     reader.set_progress_handler(Some(Box::new({
-        let mut remote_progress = progress.add_child_with_id("remote", ProgressId::RemoteProgress.into());
-        // SAFETY: Ugh, so, with current Rust I can't declare lifetimes in the involved traits the way they need to
-        //         be and I also can't use scoped threads to pump from local scopes to an Arc version that could be
-        //         used here due to the this being called from sync AND async code (and the async version doesn't work
-        //         with a surrounding `std::thread::scope()`.
-        //         Thus there is only claiming this is 'static which we know works for *our* implementations of ExtendedBufRead
-        //         and typical implementations, but of course it's possible for user code to come along and actually move this
-        //         handler into a context where it can outlive the current function. Is this going to happen? Probably not unless
-        //         somebody really wants to break it. So, with standard usage this value is never used past its actual lifetime.
-        #[allow(unsafe_code)]
-        let should_interrupt: &'static AtomicBool = unsafe { std::mem::transmute(should_interrupt) };
+        let mut remote_progress = progress.add_child_with_id("remote".to_string(), ProgressId::RemoteProgress.into());
         move |is_err: bool, data: &[u8]| {
             gix_protocol::RemoteProgress::translate_to_progress(is_err, data, &mut remote_progress);
             if should_interrupt.load(Ordering::Relaxed) {
@@ -401,5 +426,5 @@ fn setup_remote_progress<P>(
                 ProgressAction::Continue
             }
         }
-    }) as gix_protocol::transport::client::HandleProgress));
+    }) as gix_protocol::transport::client::HandleProgress<'a>));
 }
